from ._eiforest import ExtendedIsolationForest
from ._honest_forest import HonestForestClassifier
from ._multiview import MultiViewRandomForestClassifier
from ._supervised_forest import (
    ExtraObliqueRandomForestClassifier,
    ExtraObliqueRandomForestRegressor,
    ObliqueRandomForestClassifier,
    ObliqueRandomForestRegressor,
    PatchObliqueRandomForestClassifier,
    PatchObliqueRandomForestRegressor,
)
from ._unsupervised_forest import UnsupervisedObliqueRandomForest, UnsupervisedRandomForest
