from .hyppo import (
    approximate_clf_mutual_information,
    approximate_clf_mutual_information_with_monte_carlo,
    make_marron_wand_classification,
    make_quadratic_classification,
    make_trunk_classification,
    make_trunk_mixture_classification,
)
from .multiview import make_gaussian_mixture, make_joint_factor_model
